﻿unit U_IISLogParser;

////////////////////////////////////////////////////////////////////////////////////////////////////
interface

uses
  System.Classes,
  SQLiteTable3;

type
  TIISLogParser = class
  private
    FFieldMaps: TStringList;
    FDB: TSQLiteDatabase;
    FInsertEntryField: TSQLiteQuery;
    FInsertEntry: TSQLiteQuery;
    FUpdateEntry: TSQLiteQuery;
    FUpdateSection: TSQLiteQuery;
    FInsertSection: TSQLiteQuery;

    function  InsertEntry(const FileID, SectionID, LineNumber: Int64): Int64;
    procedure InsertEntryASPError(const EntryID: Int64; const SourceLine, ErrorCode, ErrorMsg: string);
    procedure InsertEntryCookie(const EntryID, Ordinal: Int64; const Name, Value: string);
    procedure InsertEntryField(const EntryID, Ordinal, FieldID: Int64; const Value: string);
    procedure InsertEntryURIParameter(const EntryID, Ordinal: Int64; const Name, Value: string);
    function  InsertFile(const FullFileName: string): Int64;
    function  InsertSection(const FileID, LineNumber: Int64): Int64;
    function  InsertSectionField(const SectionID, Ordinal: Int64; const Name: string): Int64;
    procedure InsertSectionValue(const SectionID, LineNumber: Int64; const Name, Value: string);
    procedure UpdateEntry(const EntryID: Int64; const DateTime: string; const Fields: TStrings);
    procedure UpdateSection(const SectionID: Int64;
                            const Software, Version, DateTime, Fields: string;
                            const FieldsetID: Int64);
  public
    constructor Create(const DatabaseFile: string);
    destructor  Destroy; override;

    procedure InitTables;
    function  ParseLogFile(const FileName: string): Cardinal;

    property DB: TSQLiteDatabase    read FDB;
    property FieldMaps: TStringList read FFieldMaps;
  end;

////////////////////////////////////////////////////////////////////////////////////////////////////
implementation
uses
  System.SysUtils, System.Types, System.StrUtils, System.IOUtils,
  System.NetEncoding, System.Generics.Collections, Winapi.Windows;

const
  scFieldMaps = 'cs-method=method'+sLineBreak +
                'cs-host=host'+sLineBreak +
                's-port=port'+sLineBreak +
                'cs-uri-stem=path'+sLineBreak +
                'cs-uri-query=query'+sLineBreak +
                'c-ip=remote_ip'+sLineBreak +
                'cs(User-Agent)=user_agent'+sLineBreak +
                'cs(Referer)=referrer'+sLineBreak +
                'sc-status=status'+sLineBreak +
                'sc-substatus=substatus'+sLineBreak +
                'sc-win32-status=os_status'+sLineBreak +
                'cs-bytes=request_bytes'+sLineBreak +
                'sc-bytes=response_bytes'+sLineBreak +
                'time-taken=time_taken';
  DBFS: TFormatSettings = (DateSeparator: '-'; TimeSeparator: ':';
                            ShortDateFormat: 'yyyy-MM-dd'; LongDateFormat: 'yyyy-MM-dd';
                            ShortTimeFormat: 'hh:nn:ss'; LongTimeFormat: 'hh:nn:ss';
                            ThousandSeparator: #0; DecimalSeparator: '.');

{ -------------------------------------------------------------------------------------- }
constructor TIISLogParser.Create(const DatabaseFile: string);
begin
  inherited Create;
  FDB := TSQLiteDatabase.Create(DatabaseFile);
  FDB.ExecSQL('PRAGMA foreign_keys = ON');
  FFieldMaps := TStringList.Create;
  FFieldMaps.CaseSensitive := False;
  FFieldMaps.Duplicates := dupAccept;
  FFieldMaps.Text := scFieldMaps;
  FFieldMaps.Sort;
end {TIISLogParser.Create};
{ -------------------------------------------------------------------------------------- }
destructor TIISLogParser.Destroy;
begin
  FInsertEntryField.Release;
  FInsertEntry.Release;
  FUpdateEntry.Release;
  FUpdateSection.Release;
  FInsertSection.Release;
  FDB.Free;
  FFieldMaps.Free;
  inherited;
end;

{ -------------------------------------------------------------------------------------- }
procedure TIISLogParser.InitTables;
var
  SQL: string;
  ResName: string;
  DBVersion: Integer;
begin
  DBVersion := FDB.GetTableValue('PRAGMA user_version;');
  ResName := 'create_sql_' + IntToStr(DBVersion);
  while FindResource(HInstance, PChar(ResName), RT_RCDATA) <> 0 do begin
    with TStreamReader.Create(TResourceStream.Create(HInstance, ResName, RT_RCDATA), True) do
    try
      OwnStream;
      SQL := ReadToEnd();
      while SQL <> '' do
        SQL := FDB.ExecSQL(SQL);
    finally
      Free;
    end;
    Inc(DBVersion);
    FDB.ExecSQL('PRAGMA user_version = ' + IntToStr(DBVersion));
    ResName := 'create_sql_' + IntToStr(DBVersion);
  end{while};
end {TIISLogParser.InitTables};

{ -------------------------------------------------------------------------------------- }
// http://stackoverflow.com/a/1642355/3092116
function FileSize(const aFilename: String): Int64;
var
  info: TWin32FileAttributeData;
begin
  result := -1;

  if NOT GetFileAttributesEx(PWideChar(aFileName), GetFileExInfoStandard, @info) then
    EXIT;

  Int64Rec(Result).Lo := Info.nFileSizeLow;
  Int64Rec(Result).Hi := Info.nFileSizeHigh;
end {FileSize};

{ -------------------------------------------------------------------------------------- }
function TIISLogParser.InsertFile(const FullFileName: string): Int64;
var
  LastModifiedDisk: TDateTime;
  Path, FileName: string;
  sLastModifiedDB: string;
  SizeDisk, SizeDB: Int64;
  Row: TSQLiteUniTable;
begin
  SizeDisk := FileSize(FullFileName);
  LastModifiedDisk := TFile.GetLastWriteTimeUtc(FullFileName);
  Path := ExtractFileDir(FullFileName);
  FileName := ExtractFileName(FullFileName);

  FDB.AddParamText(':path', Path);
  FDB.AddParamText(':filename', FileName);
  Row := FDB.GetUniTable('SELECT size, last_modified FROM Files WHERE path = :path AND filename = :filename');
  try
    if not Row.EOF then begin
      SizeDB := Row.FieldAsInteger(0);
      sLastModifiedDB := Row.FieldAsString(1);
    end else begin
      SizeDB := -1;
      sLastModifiedDB := '';
    end;
  finally
    Row.Free;
  end;

  if (SizeDisk = SizeDB) and (sLastModifiedDB >= DateTimeToStr(LastModifiedDisk, DBFS)) then begin
    Result := 0;
    Exit; // Leave the thing alone if it hasn't changed
  end;

  FDB.AddParamText(':path', Path);
  FDB.AddParamText(':filename', FileName);
  FDB.AddParamInt(':size', SizeDisk);
  FDB.AddParamText(':last_modified', DateTimeToStr(LastModifiedDisk, DBFS));
  FDB.ExecSQL('INSERT INTO Files (path, filename, size, last_modified) SELECT :path, :filename, :size, :last_modified');
  Result := FDB.GetLastInsertRowID;
end {TIISLogParser.InsertFile};
{ -------------------------------------------------------------------------------------- }
function TIISLogParser.InsertSection(const FileID, LineNumber: Int64): Int64;
begin
  if FInsertSection.Statement = nil then
    FInsertSection := FDB.PrepareSQL('INSERT INTO Sections (file_id, first_line) SELECT :file_id, :line');

  FDB.AddParamInt(':file_id', FileID);
  FDB.AddParamInt(':line', LineNumber);
  FDB.ExecSQL(FInsertSection);
  Result := FDB.GetLastInsertRowID;
end;
{ -------------------------------------------------------------------------------------- }
procedure TIISLogParser.InsertSectionValue(const SectionID, LineNumber: Int64; const Name, Value: string);
begin
  FDB.AddParamInt(':section_id', SectionID);
  FDB.AddParamInt(':line', LineNumber);
  FDB.AddParamText(':name', Name, '');
  FDB.AddParamText(':value', Value, '');
  FDB.ExecSQL('INSERT INTO SectionValues (section_id, line, name, value) SELECT :section_id, :line, :name, :value');
end;
{ -------------------------------------------------------------------------------------- }
function TIISLogParser.InsertSectionField(const SectionID, Ordinal: Int64; const Name: string): Int64;
begin
  FDB.AddParamInt(':section_id', SectionID);
  FDB.AddParamInt(':ordinal', Ordinal);
  FDB.AddParamText(':name', Name, '');
  FDB.ExecSQL('INSERT INTO SectionFields (section_id, ordinal, name) SELECT :section_id, :ordinal, :name');
  Result := FDB.GetLastInsertRowID;
end;
{ -------------------------------------------------------------------------------------- }
procedure TIISLogParser.UpdateSection(const SectionID: Int64;
                                      const Software, Version, DateTime, Fields: string;
                                      const FieldsetID: Int64);
begin
  if FUpdateSection.Statement = nil then
    FUpdateSection := FDB.PrepareSQL( ' UPDATE Sections' +
                                      '    SET software = :software' +
                                      '      , version = :version' +
                                      '      , datetime = :datetime' +
                                      '      , fields = :fields' +
                                      '      , fieldset_id = :fieldset_id' +
                                      ' WHERE id = :section_id');

  FDB.AddParamInt(':section_id', SectionID);
  FDB.AddParamText(':software', Software, '');
  FDB.AddParamText(':version', Version, '');
  FDB.AddParamText(':datetime', DateTime, '');
  FDB.AddParamText(':fields', Fields, '');
  FDB.AddParamInt(':fieldset_id', FieldsetID, 0);
  FDB.ExecSQL(FUpdateSection);
end {TIISLogParser.UpdateSection};
{ -------------------------------------------------------------------------------------- }
function TIISLogParser.InsertEntry(const FileID, SectionID, LineNumber: Int64): Int64;
begin
  if FInsertEntry.Statement = nil then
    FInsertEntry := FDB.PrepareSQL('INSERT INTO Entries (file_id, section_id, line) SELECT :file_id, :section_id, :line');

  FDB.AddParamInt(':file_id', FileID);
  FDB.AddParamInt(':section_id', FileID, 0);
  FDB.AddParamInt(':line', LineNumber);
  FDB.ExecSQL(FInsertEntry);
  Result := FDB.GetLastInsertRowID;
end;
{ -------------------------------------------------------------------------------------- }
procedure TIISLogParser.InsertEntryField(const EntryID, Ordinal, FieldID: Int64; const Value: string);
begin
  if FInsertEntryField.Statement = nil then
    FInsertEntryField := FDB.PrepareSQL('INSERT INTO EntryFields (entry_id, ordinal, section_field_id, value) SELECT :entry_id, :ordinal, :field_id, :value');

  FDB.AddParamInt(':entry_id', EntryID);
  FDB.AddParamInt(':ordinal', Ordinal);
  FDB.AddParamInt(':field_id', FieldID, 0);
  FDB.AddParamText(':value', Value, '');
  FDB.ExecSQL(FInsertEntryField);
end;
{ -------------------------------------------------------------------------------------- }
procedure TIISLogParser.InsertEntryASPError(const EntryID: Int64; const SourceLine, ErrorCode, ErrorMsg: string);
var
  iSourceLine: Int64;
begin
  if (SourceLine.Trim = '') and (ErrorCode.Trim = '') and (ErrorMsg.Trim = '') then
    Exit;
  FDB.AddParamInt(':entry_id', EntryID);
  if TryStrToInt64(SourceLine, iSourceLine) then
    FDB.AddParamInt(':source_line', iSourceLine)
  else
    FDB.AddParamText(':source_line', SourceLine, '');
  FDB.AddParamText(':error_code', ErrorCode, '');
  FDB.AddParamText(':message', ErrorMsg, '');
  FDB.ExecSQL('INSERT INTO EntryASPErrors (entry_id, source_line, error_code, message) SELECT :entry_id, :source_line, :error_code, :message');
end;
{ -------------------------------------------------------------------------------------- }
procedure TIISLogParser.InsertEntryURIParameter(const EntryID, Ordinal: Int64; const Name, Value: string);
begin
  FDB.AddParamInt(':entry_id', EntryID);
  FDB.AddParamInt(':ordinal', Ordinal);
  FDB.AddParamText(':name', Name, '');
  FDB.AddParamText(':value', Value, '');
  FDB.ExecSQL('INSERT INTO EntryURIParameters (entry_id, ordinal, name, value) SELECT :entry_id, :ordinal, :name, :value');
end;
{ -------------------------------------------------------------------------------------- }
procedure TIISLogParser.InsertEntryCookie(const EntryID, Ordinal: Int64; const Name, Value: string);
begin
  FDB.AddParamInt(':entry_id', EntryID);
  FDB.AddParamInt(':ordinal', Ordinal);
  FDB.AddParamText(':name', Name, '');
  FDB.AddParamText(':value', Value, '');
  FDB.ExecSQL('INSERT INTO EntryCookies (entry_id, ordinal, name, value) SELECT :entry_id, :ordinal, :name, :value');
end;
{ -------------------------------------------------------------------------------------- }
procedure TIISLogParser.UpdateEntry(const EntryID: Int64; const DateTime: string; const Fields: TStrings);
var
  i: Integer;
  ParamName, LogName: string;
begin
  if FUpdateEntry.Statement = nil then
    FUpdateEntry := FDB.PrepareSQL( ' UPDATE Entries' +
                                    '    SET datetime = :datetime' +
                                    '    ,   method = :method' +
                                    '    ,   host = :host' +
                                    '    ,   port = :port' +
                                    '    ,   path = :path' +
                                    '    ,   query = :query' +
                                    '    ,   remote_ip = :remote_ip' +
                                    '    ,   user_agent = :user_agent' +
                                    '    ,   referrer = :referrer' +
                                    '    ,   status = :status' +
                                    '    ,   substatus = :substatus' +
                                    '    ,   os_status = :os_status' +
                                    '    ,   request_bytes = :request_bytes' +
                                    '    ,   response_bytes = :response_bytes' +
                                    '    ,   time_taken = :time_taken' +
                                    '  WHERE id = :entry_id');

  FDB.AddParamInt(':entry_id', EntryID);
  FDB.AddParamText(':datetime',       DateTime, '');
  for i := 0 to FieldMaps.Count - 1 do begin
    ParamName := FieldMaps.ValueFromIndex[i];
    LogName := FieldMaps.Names[i];
    FDB.AddParamText(':'+ParamName, Fields.Values[LogName], '');
  end;
  FDB.ExecSQL(FUpdateEntry);
end {TIISLogParser.UpdateEntry};


{ -------------------------------------------------------------------------------------- }
function Unescape(const URL: string): string;
begin
  if URL = '-' then
    Result := ''
  else
    Result := TNetEncoding.URL.Decode(URL);
end {Unescape};

{ ====================================================================================== }
type
  TFieldInfo = record
    Name: string;
    ID: Int64;
    constructor Create(const Name: string; const ID: Int64);
  end;
  { ---------------------------------------------------------------------------------- }
  constructor TFieldInfo.Create(const Name: string; const ID: Int64);
  begin
    Self.Name := Name;
    Self.ID := ID;
  end;
  { ---------------------------------------------------------------------------------- }


{ -------------------------------------------------------------------------------------- }
function TIISLogParser.ParseLogFile(const FileName: string): Cardinal;
var
  EntryFields, URLParams, Cookies: TStringList;
  FileID: Int64;
  Log: TStreamReader;
  LineNumber: Int64;
  Line, Name, Value: string;
  ReadingSection: boolean;
  SectionID: Int64;
  CharPos: integer;
  Values, Elements: TArray<string>;
  Fields: TList<TFieldInfo>;
  i: integer;
  EntryID: Int64;
  FieldName: string;
  SectionSoftware, SectionVersion, SectionDate, SectionFieldNames: string;
  EntryDateTime: string;
  j: integer;
  FieldID: Int64;
  SectionFieldsID: Int64;
  SectionFields: TSQLiteUniTable;
begin
  Result := 0;
  EntryFields := TStringList.Create;
  URLParams := TStringList.Create;
  Cookies := TStringList.Create;
  Fields := TList<TFieldInfo>.Create;
  try
    URLParams.LineBreak := '&';
    URLParams.NameValueSeparator := '=';
    URLParams.Duplicates := dupAccept;

    Cookies.LineBreak := ';';
    Cookies.NameValueSeparator := '=';
    Cookies.Duplicates := dupAccept;

    FDB.SavePoint(Filename);
    try
      ReadingSection := False;
      SectionID := 0;
      SectionFieldsID := 0;
      FileID := InsertFile(FileName);
      if FileID = 0 then // we already have this file
        Exit;
      Log := TStreamReader.Create(FileName);
      try
        Fields.Clear;
        LineNumber := 0;
        repeat
          Line := Log.ReadLine;
          Inc(LineNumber);
          if Line.Length = 0 then
            Continue; // Skip empty lines
          if Line.Chars[0] = #0 then
            Break; // IIS pads its log files with null characters; this means we've reached the end of the useful log

          if Line.Chars[0] = '#' then begin
            Value := Line.Substring(1);
            // This is a section field
            if not ReadingSection then begin
              SectionID := InsertSection(FileID, LineNumber);
              ReadingSection := True;
              SectionSoftware := '';
              SectionVersion := '';
              SectionDate := '';
              SectionFieldNames := '';
              SectionFieldsID := 0;
            end;
            CharPos := Value.IndexOf(':');
            if CharPos >= 0 then begin
              Name := Value.Substring(0, CharPos).Trim;
              Value := Value.Substring(CharPos + 1).Trim;
            end else begin
              Name := Value;
              Value := '';
            end;
            // Now check for known fields
            case IndexText(Name, ['software', 'version', 'date', 'fields']) of
              0: SectionSoftware := Value;
              1: SectionVersion := Value;
              2: SectionDate := Value;
              3: begin // fields
                SectionFieldNames := Value;
                FDB.AddParamText(':fields', SectionFieldNames);
                SectionFieldsID := FDB.GetTableValue('SELECT id FROM Sections WHERE fields = :fields');
                if SectionFieldsID = 0 then begin
                  Values := Value.Split([' ']);
                  for i := Low(Values) to High(Values) do begin
                    FieldID := InsertSectionField(SectionID, i, Values[i]);
                    Fields.Add(TFieldInfo.Create(Values[i], FieldID));
                  end{for};
                end else begin
                  SectionFieldNames := '';
                  FDB.AddParamInt(':fieldset_id', SectionFieldsID);
                  SectionFields := FDB.GetUniTable('SELECT name, id FROM SectionFields WHERE section_id = :fieldset_id ORDER BY ordinal');
                  try
                    while not SectionFields.EOF do begin
                      Fields.Add(TFieldInfo.Create(SectionFields.FieldAsString(0), SectionFields.FieldAsInteger(1)));
                      SectionFields.Next;
                    end;
                  finally
                    SectionFields.Free;
                  end;
                end;
              end {case 'fields'};
              else begin
                // All other cases need to be stored separately
                InsertSectionValue(SectionID, LineNumber, Name, Value);
              end;
            end{case};

          end else begin

            // This is a regular line, not a section field
            if ReadingSection then begin
              // If we were reading a section, then wrap that up.
              if (SectionID > 0) then begin
                UpdateSection(SectionID, SectionSoftware, SectionVersion, SectionDate, SectionFieldNames, SectionFieldsID);
              end;
              ReadingSection := False;
            end;

            EntryDateTime := '';
            EntryFields.Clear;
            URLParams.Clear;
            EntryID := InsertEntry(FileID, SectionID, LineNumber);
            Values := Line.Split([' ']);
            for i := Low(Values) to High(Values) do begin
              Value := Unescape(Values[i]);
              if Value = '-' then
                Value := '';
              if i < Fields.Count then begin
                FieldName := Fields[i].Name;
                FieldID := Fields[i].ID;
              end else begin
                FieldName := '';
                FieldID := 0;
              end;
              // Now check for known fields
              case IndexText(FieldName, ['date', 'time', 'cs-uri-query', 'cs(Cookie)']) of
                0: EntryDateTime := (Value + ' ' + EntryDateTime).Trim;
                1: EntryDateTime := (EntryDateTime + ' ' + Value).Trim;
                2: begin // cs-uri-query
                  Value := Values[i]; // we need the unescaped version for this one
                  Elements := Value.Split(['|']);
                  if Length(Elements) > 0 then begin
                    Value := Elements[0];
                    if Length(Elements) > 1 then begin
                      // TODO: unescape each element
                      // TODO: NULL if an element doesn't exist
                      InsertEntryASPError(EntryID, Unescape(Elements[1]),
                                                    Unescape(Elements[2]),
                                                    Unescape(Elements[3]));
                    end;
                  end;
                  if Value = '-' then
                    Value := '';
                  EntryFields.Values[FieldName] := Value;
                  if Value <> '' then begin
                    URLParams.Text := Value;
                    for j := 0 to URLParams.Count - 1 do begin
                      if URLParams[j].IndexOf(URLParams.NameValueSeparator) = -1 then begin
                        Name := Unescape(URLParams[j]);
                        Value := '';
                      end else begin
                        Name := Unescape(URLParams.Names[j]);
                        Value := Unescape(URLParams.ValueFromIndex[j]);
                      end;
                      InsertEntryURIParameter(EntryID, j, Name, Value);
                    end {for};
                  end;
                end {case cs-query};
                3: begin // cs(Cookie)
                  Value := Values[i]; // use the undecoded value
                  if Value = '-' then
                    Value := '';
                  Cookies.Text := Value;
                  for j := 0 to Cookies.Count - 1 do begin
                    if Cookies[j].IndexOf(Cookies.NameValueSeparator) = -1 then begin
                      Name := Unescape(Cookies[j]).Trim;
                      Value := '';
                    end else begin
                      Name := Unescape(Cookies.Names[j]).Trim;
                      Value := Unescape(Cookies.ValueFromIndex[j]).Trim;
                    end;
                    InsertEntryCookie(EntryID, j, Name, Value);
                  end {for};
                end {case cs(Cookie)};
                else begin
                  EntryFields.Values[FieldName] := Value;
                  if FieldMaps.IndexOfName(FieldName) = -1 then
                    InsertEntryField(EntryID, i, FieldID, Value);
                end;
              end{case};
            end{for};
            UpdateEntry(EntryID, EntryDateTime, EntryFields);
            Inc(Result);
          end{if};

        until Log.EndOfStream;
      finally
        Log.Free;
      end;
    finally
      FDB.Release(FileName);
    end;
  finally
    Cookies.Free;
    URLParams.Free;
    EntryFields.Free;
    Fields.Free;
  end;
end {TIISLogParser.ParseLogFile};

end.
